import ast
import configparser
import glob
import json
import logging
import os
import random
from dataclasses import InitVar, dataclass, field
from typing import List, Tuple

import numpy as np
import pandas as pd
from alive_progress import alive_bar

from database_tools.io.data import write_to_json
from database_tools.io.utils import load_valid_records
from database_tools.io.wfdb import get_data_record, get_signal
from database_tools.processing.detect import (detect_flat_lines,
                                              detect_notches, detect_peaks)
from database_tools.processing.metrics import (get_beat_similarity,
                                               get_similarity, get_snr)
from database_tools.processing.modify import align_signals, bandpass
from database_tools.processing.utils import repair_peaks_troughs_idx, window

log_dir = '/'.join(os.getcwd().split('/')[0:5]) + '/logs/'
if 'database_tools' in log_dir:
    logging.basicConfig(
        filename=f'{log_dir}tools.log',
        format= '[%(asctime)s] {%(pathname)s:%(lineno)d} %(levelname)s - %(message)s',
        datefmt='%H:%M:%S'
    )
    logger = logging.getLogger('tools')
    logger.setLevel(logging.INFO)


@dataclass(frozen=True)
class ConfigItem:
    def set_attr(self, config_item):
        for key, value in config_item.items():
            try:
                object.__setattr__(self, key, ast.literal_eval(value))
            except:
                object.__setattr__(self, key, value)


@dataclass(frozen=True)
class ConfigMapper:

    config_path: InitVar[str]
    data = ConfigItem()
    train = ConfigItem()
    model = ConfigItem()
    sweep = ConfigItem()
    deploy = ConfigItem()

    def __post_init__(self, config_path: str):
        config = configparser.ConfigParser()
        config.read(config_path)
        self.data.set_attr(config['DATASET_PIPELINE'])
        self.train.set_attr(config['TRAINING_PIPELINE'])
        self.model.set_attr(config['MODEL'])
        self.sweep.set_attr(config['SWEEP'])
        self.deploy.set_attr(config['DEPLOYMENT'])


@dataclass
class MetricLogger:

    cm: InitVar[ConfigMapper]
    metrics_dict: dict = field(init=False)
    valid_samples: int = 0
    rejected_samples: int = 0
    patient_samples: int = 0
    prev_patient: str = ''

    def __post_init__(self, cm: list):
        self.metrics_dict = {}
        for m in cm.metrics:
            self.metrics_dict.update({m: []})
        self.samples_per_file = cm.samples_per_file
        self.samples_per_patient = cm.samples_per_patient
        self.max_samples = cm.max_samples

    def update_metrics(self, metrics: dict) -> None:
        """Update dataset metrics dictionary and return next action.
           0 -> Continue
           1 -> Write data to file and continue.
           2 -> Write remaining data and stop.

        Args:
            metrics (dict): _description_

        Returns:
            _type_: _description_
        """
        patient = metrics['record_path'].split('/')[1]
        if self.prev_patient != patient:
            self.patient_samples = 0
            self.prev_patient = patient
        if metrics['valid']:
            self.valid_samples += 1
            self.patient_samples += 1
        else:
            self.rejected_samples += 1

        for key in self.metrics_dict.keys():
            self.metrics_dict[key].append(metrics[key])

        if (self.valid_samples % self.samples_per_file) == 0:
            if self.valid_samples >= self.max_samples:
                return 2
            else:
                return 1
        else:
            return 0

    def save_stats(self, file_path):
        df = pd.DataFrame(self.metrics_dict)
        df.to_csv(file_path, index=False)
        print(f'Saving dataset stats to {file_path}')


@dataclass
class Dataset:

    data_dir: InitVar[str]
    ppg: np.ndarray = field(init=False)
    vpg: np.ndarray = field(init=False)
    apg: np.ndarray = field(init=False)
    abp: np.ndarray = field(init=False)

    _nframes: int = field(init=False)

    def __post_init__(self, data_dir: str) -> None:
        frames = [pd.read_json(file, lines=True) for file in glob.glob(f'{data_dir}data/lines/*.jsonlines')]
        object.__setattr__(self, '_nframes', len(frames))
        data = pd.concat(frames, ignore_index=True)
        object.__setattr__(self, 'ppg', np.array(data['ppg'].to_list()))
        vpg = np.gradient(self.ppg, axis=1)  # 1st derivative of ppg
        apg = np.gradient(vpg, axis=1) # 2nd derivative of vpg
        object.__setattr__(self, 'vpg', vpg)
        object.__setattr__(self, 'apg', apg)
        object.__setattr__(self, 'abp', np.array(data['abp'].to_list()))
        self.info

    @property
    def info(self):
        logger.info(f'Data was extracted from {self._nframes} JSONLINES files.')
        logger.info(f'The total number of windows is {self.ppg.shape[0]}.')


@dataclass
class Window:

    sig: np.ndarray
    cm: ConfigMapper
    checks: List[str]

    @property
    def _snr_check(self) -> bool:
        self.snr, self.f0 = get_snr(self.sig, low=self.cm.freq_band[0], high=self.cm.freq_band[1], df=0.2, fs=self.cm.fs)
        return self.snr > self.cm.snr

    @property
    def _hr_check(self) -> bool:
        return (self.f0 > self.cm.hr_freq_band[0]) & (self.f0 < self.cm.hr_freq_band[1])

    @property
    def _flat_check(self) -> bool:
        return not detect_flat_lines(self.sig, n=self.cm.flat_line_length)

    @property
    def _beat_check(self) -> bool:
        self.beat_sim = get_beat_similarity(
            self.sig,
            troughs=self.troughs,
            fs=self.cm.fs,
        )
        return self.beat_sim > self.cm.beat_sim

    @property
    def _notch_check(self) -> bool:
        notches = detect_notches(
            self.sig,
            peaks=self.peaks,
            troughs=self.troughs,
        )
        return len(notches) > self.cm.min_notches

    @property
    def _bp_check(self) -> bool:
        self.dbp, self.sbp = np.min(self.sig), np.max(self.sig)
        dbp_check = (self.dbp > self.cm.dbp_bounds[0]) & (self.dbp < self.cm.dbp_bounds[1])
        sbp_check = (self.sbp > self.cm.sbp_bounds[0]) & (self.sbp < self.cm.sbp_bounds[1])
        return dbp_check & sbp_check

    @property
    def valid(self) -> bool:
        v = [object.__getattribute__(self, '_' + c + '_check') for c in self.checks]
        return np.array(v).all()

    def get_peaks(self, pad_width=40) -> None:
        x_pad = np.pad(self.sig, pad_width=pad_width, constant_values=np.mean(self.sig))
        peaks, troughs = detect_peaks(x_pad).values()
        peaks, troughs = repair_peaks_troughs_idx(peaks, troughs)
        self.peaks = peaks - pad_width - 1
        self.troughs = troughs - pad_width - 1


def congruency_check(ppg: Window, abp: Window, cm: ConfigMapper) -> Tuple[float, float, bool]:
    """Performs checks between ppg and abp Window objects.

    Args:
        ppg (Window): Object with ppg data.
        abp (Window): Object with abp data.
        cm (ConfigMapper): Config mapping dataclass.

    Returns:
        bool: True if valid, False if not.
    """
    time_sim = get_similarity(ppg.sig, abp.sig)
    ppg_f = np.abs(np.fft.fft(ppg.sig))
    abp_f = np.abs(np.fft.fft(bandpass(abp.sig, low=cm.freq_band[0], high=cm.freq_band[1], fs=cm.fs)))
    spec_sim = get_similarity(ppg_f, abp_f)
    sim_check = (time_sim > cm.sim) & (spec_sim > cm.sim)
    hr_delta_check = np.abs(ppg.f0 - abp.f0) < cm.hr_delta
    congruency_check = sim_check & hr_delta_check
    return {
        'time_sim': time_sim,
        'spec_sim': spec_sim,
        'cong_check': congruency_check
    }


class DatasetFactory():
    def __init__(
        self,
        data_dir,
    ) -> None:
        self._data_dir = data_dir
        self._partner = 'mimic3'
    def run(self):
        cm = ConfigMapper(self._data_dir + '/config.ini').data
        metrics_logger = MetricLogger(cm)

        print('Gettings valid segments...')
        valid_records = load_valid_records(self._data_dir + 'valid_records.csv')

        print('Collecting samples...')
        json_data = ''
        for record_path in valid_records:
            data = self._get_data(record_path)
            if data:
                ppg, abp = data[0], data[1]
            else:
                continue

            # Apply bandpass filter to ppg
            ppg = bandpass(ppg, low=cm.freq_band[0], high=cm.freq_band[1], fs=cm.fs)

            overlap = int(cm.fs / 2)
            l = cm.win_len + overlap
            idx = window(ppg, l, overlap)

            title_len = 50
            with alive_bar(total=len(idx), title_length=title_len, force_tty=True) as bar:
                for i, j in idx:
                    bar.title(f'Rejected: {metrics_logger.rejected_samples} --- Collected: {metrics_logger.valid_samples}'.rjust(title_len))
                    p = ppg[i:j]
                    a = abp[i:j]

                    p, a = align_signals(p, a, win_len=cm.win_len, fs=cm.fs)

                    # Evaluate ppg
                    p_win = Window(p, cm, cm.checks)
                    p_win.get_peaks()
                    p_valid = p_win.valid

                    # Evaluate abp
                    a_win = Window(a, cm, cm.checks + ['bp'])
                    a_win.get_peaks()
                    a_valid = a_win.valid

                    # Evaluate ppg vs abp
                    congruence_check = congruency_check(p_win, a_win, cm)
                    valid = p_valid & a_valid & congruence_check['cong_check']

                    metrics = self._compile_window_metrics(record_path, (i, j), valid, p_win, a_win, congruence_check)
                    status = metrics_logger.update_metrics(metrics)

                    if valid:
                        json_data += json.dumps(dict(ppg=p.tolist(), abp=a.tolist())) + '\n'

                        # Write to file when count is reached. 
                        if status in [1, 2]:
                            file_name = str(int(metrics_logger.valid_samples / cm.samples_per_file) - 1).zfill(3)
                            file_path = self._data_dir + f'data/lines/{self._partner}_{file_name}.jsonlines'
                            write_to_json(json_data, file_path); json_data = ''
                            if status == 2:
                                metrics_logger.save_stats(self._data_dir + f'{self._partner}_stats.csv')
                                print('Done!')
                                return
                    bar()
                    if metrics_logger.patient_samples == cm.samples_per_patient:
                        break
        return

    def _get_valid_segs(self, valid_path):
        valid_df = pd.read_csv(valid_path)

        # Extract data from valid_df and shuffle in the same order
        temp = list(zip(valid_df['id'], valid_df['url']))
        random.shuffle(temp)
        patient_ids, files = [list(t) for t in zip(*temp)]
        patient_ids, files = list(patient_ids), list(files)
        return (patient_ids, files)

    def _get_data(self, path):
        rec = get_data_record(path=path, record_type='waveforms')
        if rec is not None:
            ppg = get_signal(rec, sig='PLETH')
            abp = get_signal(rec, sig='ABP')
            ppg[np.isnan(ppg)] = 0
            abp[np.isnan(abp)] = 0
            return (ppg, abp)
        else:
            return False

    def _compile_window_metrics(self, record_path: str, window_idx: Tuple[int, int], valid, ppg: Window, abp: Window, congruence_check: dict) -> dict:
        metrics = {
            'record_path': record_path,
            'window_idx': window_idx,
            'valid': valid,
            'time_sim': congruence_check['time_sim'],
            'spec_sim': congruence_check['spec_sim'],
            'ppg_snr': ppg.snr,
            'abp_snr': abp.snr,
            'ppg_hr': ppg.f0 * 60,
            'abp_hr': abp.f0 * 60,
            'sbp': abp.sbp,
            'dbp': abp.dbp,
            'ppg_beat_sim': ppg.beat_sim,
            'abp_beat_sim': abp.beat_sim,
            'flat_ppg': ppg._flat_check,
            'flat_abp': abp._flat_check,
            'ppg_notches': ppg._notch_check,
            'abp_notches': abp._notch_check,
        }
        return metrics



def write_to_json(data: str, file_path: str) -> None:
    """Writes data to .json or .jsonlines file. 

    Args:
        data (str): JSON or JSONLINES formatted string.
        file_path (str): Path to file.
    """
    if (file_path.split('.') == '.jsonlines') & ('\n' not in data):
        raise ValueError('Data must contain new lines to be written to a .jsonlines file.')
    with open(file_path, 'w') as f:
        f.write(data)


class TFRecordExamples:
    pass



import numpy as np
from scipy import signal
from database_tools.processing.metrics import get_similarity

def align_signals(ppg, abp, win_len, fs):
    """
    Find the index at which the two signals
    have the largest time correlation.

    Args:
        ppg (np.ndarray): PPG data.
        abp (np.ndarray): ABP data.
        win_len (int): Length of windows.
        fs (int, optional): Sampling rate of signal.

    Returns:
        signals (tuple(np.ndarray)): Aligned PPG and ABP window.
    """
    max_offset = int(fs / 2)

    abp = abp[0:win_len]

    corr = []
    for offset in range(0, max_offset):
        x = ppg[offset : win_len + offset]
        corr.append(get_similarity(x, abp))
    idx = np.argmax(corr)
    ppg_shift = ppg[idx : win_len + idx]
    return (ppg_shift, abp)

def bandpass(x, low, high, fs, method='cheby2'):
    """
    Apply one of the following bandpass filters.
      - 4th order Cheby II filter.
      - 4th order butterworth filter.

    Args:
        x (np.ndarray): Signal data.
        low (float, optional): Lower frequency in Hz.
        high (float, optional): Upper frequency in Hz.
        fs (int, optional): Sampling rate.
        method (str, optional): One of ['cheby2', 'butter']. Defaults to 'cheby2'.

    Returns:
        x (np.ndarray): Filtered signal.
    """
    if method == 'cheby2':
        filt = signal.cheby2(
            N=4,
            rs=20,
            Wn=[low, high],
            btype='bandpass',
            output='sos',
            fs=fs,
        )
    elif method == 'butter':
        filt = signal.butter(
            4,
            [low, high],
            btype='bandpass',
            output='sos',
            fs=fs
        )
    else:
        raise ValueError('Method must be one of [\'cheby2\', \'butter\']')
    x = signal.sosfiltfilt(filt, x, padtype=None)
    return x



import wfdb
from dataclasses import dataclass, InitVar, field
from database_tools.datastores.signals import SignalStore, SignalGroup


@dataclass
class WaveformRecord:
    """Interface for wfdb waveforms records.
       Makes it easier to peak at data and build signal
       processing pipelines.

    Attributes
    ----------
    name (str): Record name.
    fs (int): Sampling rate of wfdb.Record.
    """

    rcd: InitVar[wfdb.Record]
    name: str = field(init=False)
    fs: int = field(init=False)

    def __post_init__(self, rcd: wfdb.Record):
        setattr(self, 'name', rcd.record_name)
        setattr(self, 'fs', rcd.fs)
        signals = {sig: SignalStore(rcd.p_signal[:, i], rcd.fmt[i]) for i, sig in enumerate(rcd.sig_name) if sig in ['PLETH', 'ABP']}
        group = SignalGroup(signals)
        setattr(self, 'sigs', group)


@dataclass
class NumericsRecord:
    def __post_init__(self):
        #TODO:
        pass



import itertools
import numpy as np
from scipy import signal, integrate
from database_tools.processing.utils import make_equal_len

def get_similarity(x, y):
    """
    Calculates time or spectral similarity.

    Args:
        x (np.ndarray): PPG data.
        y (np.ndarray): ABP data.
        spectral (boolean): If True calculate fft
            of signals.

    Returns:
        coef (float): Pearson correlation coefficient.
    """
    x_bar = np.mean(x)
    y_bar = np.mean(y)
    y_temp = y - y_bar
    x_temp = x - x_bar
    covar = np.sum( (x_temp * y_temp) )
    var = np.sqrt( np.sum( (x_temp ** 2 ) ) * np.sum( (y_temp ** 2) ) )
    with np.errstate(divide='ignore', invalid='ignore'):
        try:
            coef = covar / var
        except (ZeroDivisionError, RuntimeWarning):
            coef = 0
    return coef

def get_snr(x, low, high, df, fs):
    """
    Calculate the Signal-to-noise ratio (SNR) of the cardiac signal.
    Density of spectrum between low and high frequencies is considered
    signal power. Density of spectrum outside low to high frequency
    band is considered signal noise. F0 is estimated to be the frequency
    at which the power spectrum is at its maximum.

    Args:
        x (np.ndarray): Cardiac signal data.
        low (float): Lower frequency in Hz.
        high (float): Upper frequency in Hz.
        df (float): Delta f for calculating power.
        fs (int): Sampling rate of signal.

    Returns:
        snr (float): SNR of signal in dB.
        f0 (float): Fundamental frequency of signal in Hz.
    """
    # Estimate spectral power density
    freqs, psd = signal.periodogram(x, fs, nfft=2048)
    freq_res = freqs[1] - freqs[0]

    dpdf = np.diff(psd) / np.diff(freqs)
    peak_idx = np.where(np.diff(np.sign(dpdf)))[0] + 1
    peak_x = freqs[peak_idx]
    try:
        f_idx = np.sort(np.argpartition(psd[peak_idx], -3)[-3:])
        f0, f1, f2 = peak_x[f_idx]
    except:
        return -10, 0

    # Signal power
    idx_sig_fund = np.logical_and(freqs >= f0 - df, freqs <= f0 + df)
    idx_sig_harm1 = np.logical_and(freqs >= f1 - df, freqs <= f1 + df)
    idx_sig_harm2 = np.logical_and(freqs >= f2 - df, freqs <= f2 + df)

    # Try statements are easiest way of preventing errors during integration
    try:
        p_sig_fund = integrate.simps(psd[idx_sig_fund], dx=freq_res)
    except:
        p_sig_fund = 0
    try:
        p_sig_harm1 = integrate.simps(psd[idx_sig_harm1], dx=freq_res)
    except:
        p_sig_harm1 = 0
    try:
        p_sig_harm2 = integrate.simps(psd[idx_sig_harm2], dx=freq_res)
    except:
        p_sig_harm2 = 0
    p_sig = p_sig_fund + p_sig_harm1 + p_sig_harm2

    # Noise power
    idx_cardiac = np.logical_and(freqs >= low, freqs <= high)

    try:
        p_cardiac = integrate.simps(psd[idx_cardiac], dx=freq_res)
    except:
        p_cardiac = 0
    p_noise = p_cardiac - p_sig

    # Try, except to prevent divide by 0 error
    with np.errstate(divide='ignore', invalid='ignore'):
        try:
            # Find SNR and convert to dB
            snr = 10 * np.log10(p_sig / p_noise)
        except (ZeroDivisionError, RuntimeWarning):
            snr = -10
    return snr, f0

def get_beat_similarity(x, troughs, fs, return_beats=False):
    """Calculates beat similarity by segmenting beats at valleys and
       calculating the Pearson correlation coefficient. The final value
       output is the mean of the correlation coefficients calculated from
       every combination (n=2) of beats in the signal.

       Returns -1 if there are less than 2 beats.
       ZeroDivisionError returns -2.

    Args:
        x (np.ndarray): Cardiac signal.
        troughs (list): Trough indices.
        fs (int): Sampling rate of cardiac signal.

    Returns:
        s (float): Mean of beat correlation coeffients.
    """
    # check there are at least 2 peaks
    if (len(troughs) < 2):
        return -1

    neg_len = lambda x : len(x) * -1
    beats = sorted(np.split(x, troughs), key=neg_len)

    aligned_beats = [beats[0]]

    for i, b in enumerate(beats[1::]):
        b_new = np.pad(b, pad_width=[0, len(beats[0]) - len(b)])
        aligned_beats.append(b_new)

    aligned_beats = [b for b in aligned_beats if len(b[b != 0]) > fs / 2]
    if return_beats:
        return aligned_beats

    # Get a list of every combination of beats (ie 3 beats -> [[0, 1], [0, 2], [1, 2]])
    idx = [(i, j) for ((i, _), (j, _)) in itertools.combinations(enumerate([i for i in range(len(aligned_beats))]), 2)]

    s = 0
    for i, j in idx:
        beat1, beat2 = make_equal_len(aligned_beats[i], aligned_beats[j])
        s += get_similarity(beat1, beat2)
    try:
        return s / len(aligned_beats)
    except ZeroDivisionError:
        return -2



import numpy as np
from neurokit2.ppg.ppg_findpeaks import _ppg_findpeaks_bishop

def estimate_pulse_rate(ppg_red, ppg_ir, red_idx, ir_idx, fs):
    pulse_rate_multiplier = 60 / (len(ppg_red) / fs)
    pulse_rate = int(len(red_idx['peaks']) * pulse_rate_multiplier)
    return pulse_rate

def estimate_spo2(ppg_red: list, ppg_ir: list, red_idx, ir_idx):
    """Estimate absorbtion and SpO2.

    Args:
        ppg_red (list): PPG data (red LED).
        ppg_ir (list): PPG data (infrared LED).
        red_idx (dict): Peak data for ppg_red.
        ir_idx (dict): Peak data for ppg_ir.

    Returns:
        spo2 (float): SpO2 as a percentage.
        r (float): Absorption.
    """
    ppg_red = np.array(ppg_red)
    ppg_ir = np.array(ppg_ir)
    red_peaks, red_troughs = red_idx['peaks'], red_idx['troughs']
    red_high, red_low = np.max(ppg_red[red_peaks]), np.min(ppg_red[red_troughs])

    ir_peaks, ir_troughs = ir_idx['peaks'], ir_idx['troughs']
    ir_high, ir_low = np.max(ppg_ir[ir_peaks]), np.min(ppg_ir[ir_troughs])

    ac_red = red_high - red_low
    ac_ir = ir_high - ir_low

    r = ( ac_red / red_low ) / ( ac_ir / ir_low )
    spo2 = round(104 - (17 * r), 1)  # round to 1 decimal place
    return (spo2, r)



import os
import pandas as pd
from tqdm import tqdm
from shutil import rmtree
from wfdb import rdheader


class MimicDataLocator():
    def __init__(
        self,
        data_dir='/path/to/data/dir/',
        mimic3_dir='physionet.org/files/mimic3wdb/1.0/',
    ) -> None:
        self._data_dir = data_dir
        self._mimic3_dir = mimic3_dir

    def run(self):
        records_path = self._mimic3_dir + 'RECORDS-adults'
        os.system(f'wget -q -np https://{records_path} -P {self._data_dir}')

        patient_ids, valid_segs = [], []
        for i, folder in tqdm(enumerate(self._mrn_generator(self._data_dir + 'RECORDS-adults')), total=59344):
            layout = self._get_layout(folder)
            if not layout:
                continue
            if not self._patient_has_pleth_abp(layout):
                continue
            master_header = self._get_master_header(folder)
            if not master_header:
                continue
            temp = self._valid_segments(folder, master_header)
            patient_ids += [folder.split('/')[1]] * len(temp)
            valid_segs += temp
            if (i % 100) == 0:
                rmtree(self._mimic3_dir, ignore_errors=True)

        df = pd.DataFrame(dict(id=patient_ids, url=valid_segs))
        df.to_csv(self._data_dir + 'valid_segs.csv', index=False)
        return

    def _mrn_generator(self, path):
        records = set(pd.read_csv(path, names=['records'])['records'])
        for folder in records:
            yield folder

    def _download(self, path):
        response = os.system(f'wget -q -r -np https://{path}')
        return response

    def _get_layout(self, folder):
        file = folder.split('/')[1] + '_layout'
        path = self._mimic3_dir + folder + file
        response = self._download(path + '.hea')
        if response == 0:
            layout = rdheader(path)
            return layout
        return False

    def _patient_has_pleth_abp(self, layout):
        if (layout is None) | (layout.sig_name is None):
            return False
        elif ('PLETH' in layout.sig_name) & ('ABP' in layout.sig_name):
            return True
        return False

    def _get_master_header(self, folder):
        file = folder.split('/')[1]
        path = self._mimic3_dir + folder + file
        response = self._download(path + '.hea')
        if response == 0:
            master_header = rdheader(path)
            return master_header
        return False

    def _valid_segments(self, folder, master_header):
        seg_name = master_header.seg_name
        seg_len = master_header.seg_len

        segments = []
        for name, n_samples in zip(seg_name, seg_len):
            if (n_samples > 75000) & (name != '~'):
                path = self._mimic3_dir + folder + name
                response = self._download(path + '.hea')
                if response == 0:
                    hea = rdheader(path)
                    if ('PLETH' in hea.sig_name) & ('ABP' in hea.sig_name):
                        segments.append('https://' + path)
        return segments



import random

def load_valid_records(file_path: str, shuffle: bool = True) -> list:
    with open(file_path, 'r') as f:
        vr = [line.strip('\n') for line in f]
    if shuffle:
        random.shuffle(vr)
    return vr



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['database_tools',
 'database_tools.datastores',
 'database_tools.io',
 'database_tools.processing',
 'database_tools.tools']

package_data = \
{'': ['*']}

install_requires = \
['alive-progress>=3.0.1,<4.0.0',
 'heartpy>=1.2.7,<2.0.0',
 'neurokit2>=0.2.1,<0.3.0',
 'numpy>=1.23.4,<2.0.0',
 'openpyxl>=3.0.10,<4.0.0',
 'pandas>=1.5.1,<2.0.0',
 'plotly>=5.10.0,<6.0.0',
 'pytest>=7.2.1,<8.0.0',
 'scikit-learn>=1.1.2,<2.0.0',
 'scipy>=1.9.3,<2.0.0',
 'tabulate>=0.9.0,<0.10.0',
 'tensorflow==2.9.2',
 'tqdm>=4.64.1,<5.0.0',
 'vitaldb>=1.2.10,<2.0.0',
 'wfdb>=4.0.0,<5.0.0']

setup_kwargs = {
    'name': 'heartfelt-tools',
    'version': '0.9.0',
    'description': 'Extract and process photoplethysmography and arterial blood pressure data from mimic3-waveforms and vitaldb.',
    'long_description': '<!-- Improved compatibility of back to top link: See: https://github.com/othneildrew/Best-README-Template/pull/73 -->\n<a name="readme-top"></a>\n\n<!-- PROJECT LOGO -->\n<br />\n<div align="center">\n      \n            \n <img src="images/heartfelt-logo.png"  width="300em" height="300em"> \n\n  \n  <h1 align="center">MIMIC-III Database Tools</h1>\n\n  <p align="center">\n    For extracting and cleaning ppg and abp data from the MIMIC-III Waveforms Database.\n    <br />\n  </p>\n</div>\n\n\n\n<!-- TABLE OF CONTENTS -->\n<details>\n  <summary>Table of Contents</summary>\n  <ol>\n    <li>\n      <a href="#introduction">Introduction</a>\n    </li>\n    <li>\n      <a href="#getting-started">Getting Started</a>\n      <ul>\n        <li><a href="#prerequisites">Prerequisites</a></li>\n      </ul>\n    </li>\n    <li><a href="#usage">Usage</a></li>\n      <ul>\n        <li><a href="#poetry">Poetry</a></li>\n        <li><a href="#get-valid-records">Get Valid Records</a></li>\n        <li><a href="#build-database">Build Database</a></li>\n        <li><a href="#evaluate-dataset">Evaluate Dataset</a></li>\n        <li><a href="#generate-records">Generate Records</a></li>\n        <li><a href="#read-records">Read Records</a></li>\n      </ul>\n    <li><a href="#license">License</a></li>\n  </ol>\n</details>\n\n\n\n<!-- Introduction -->\n## Introduction\n\nThis repo contains a set of tools for extracting and cleaning photoplethysmography (ppg) and artial blood pressure (abp) waveforms from the [MIMIC-III Waveforms Database](https://physionet.org/content/mimic3wdb/1.0/) for the purpose of blood pressure estimation via deep learning. \n\n<p align="right">(<a href="#readme-top">back to top</a>)</p>\n\n<!-- GETTING STARTED -->\n## Getting Started\n\nThis sections details the requirements to start using this library. Links are for Ubuntu installation.\n\n### Prerequisites\n\n1. Python\n```shell\nsudo apt install python3.8 -y\nsudo apt install python3.8-dev python3.8-venv -y\n\necho \'export PATH="$PATH:/home/ubuntu/.local/bin"\' >> ~/.bashrc\nsource ~/.bashrc\n\ncurl -sS https://bootstrap.pypa.io/get-pip.py | python3.8\npython3.8 -m pip install virtualenv\npython3.8 -m venv .venv/base-env\necho \'alias base-env="source ~/.venv/base-env/bin/activate"\' >> ~/.bashrc\nbase-env\n\npython3.8 -m pip install --upgrade pip\n```\n2. Poetry\n```shell\ncurl -sSL https://install.python-poetry.org | python3 -\necho \'export PATH="$PATH:$HOME/.local/bin"\' >> ~/.bashrc\nsource ~/.bashrc\n\n# Verify installation\npoetry --version\n```\n\n<p align="right">(<a href="#readme-top">back to top</a>)</p>\n\n\n\n<!-- USAGE EXAMPLES -->\n## Usage\n\n### Poetry\nThe commands below can be used to install the poetry environment, build the project, and activate the environment.\n```shell\ncd database-tools\npoetry lock\npoetry install\npoetry build\npoetry shell\n```\n\n### Create Data Directory\nThe functions in this library rely on a data folder named with the convention `data-YYYY-MM-DD`. This directory contains two additional folders, `mimic3/` and `figures/`. The `mimic3/lines/` folder is intended to hold the jsonlines files the data will initially saved to. The `mimic3/records/` folder will hold the TFRecords files generated from these jsonlines files. This will be discussed in greater depth in the <a href="#generate-records">Generate Records</a> section.\n\n### Get Valid Records\nThe class DataLocator (located in `database_tools/tools/`) is specifically written to find valid data files in the MIMIC-III Waveforms subset and create a csv of the html links for these data files. Performing this task prior to downloading is done to improve runtime and the usability of this workflow. Valid records refers to data files that contain both PPG and ABP recordings and are at least 10 minutes in length. Currently this code is only intended for the MIMIC-III Waveforms subset but will likely be adapated to allow for valid segments to be identified in the MIMIC-III Matched Subset (records are linked to clinical data). To perform an extraction the file `scripts/get-valid-segs.py` can be run (data directory and repository path must be configured manually). This function will output a csv called `valid-segments.csv` to the data directory provided. The figure below shows how these signals are located.\n\nAdd mimic3 valid segs logic figure.\n\n### Build Database\nThe class `BuildDatabase` (located in `database_tools/tools/`) downloads data from `valid-segments.csv`, extracts PPG and ABP data, and then processed it by leveraging the `SignalProcessor` class (located in `database_tools/preprocessing/`). A database can be build by running `scripts/build_database.py` (be sure to configure the paths). BuildDatabase takes a few important parameters which modify how signals are excluded and how the signals are divided prior to processing. The `win_len` parameter controls the length of each window, `fs` is the sampling rate of the data (125 Hz in the case of MIMIC-III), while `samples_per_file`, `samples_per_patient`, and `max_samples` control the size of the dataset (how many files the data is spread across, how many samples a patient can contribute, and the total number of samples in the dataset. The final parameter `config` controls the various constants of the SignalProcessor that determine the quality threshold for accepting signals. The SignalProcessor filters signals according to the figure chart below. The functions used for this filtering can be found in `database_tools/preprocessing/`. Data exctracted with this script is saved directly to the `mimic3/lines/` folder in the data directory. A file named `mimic3_stats.csv` containing the stats of every processed waveform (not just the valid ones) will also be saved to the data directory.\n\nAdd data preprocessing figure.\n\n### Evaluate Dataset\nThe class `DataEvaluator` (located in `database_tools/tools/`) reads the `mimic3_stats.csv` file from the provided data directory and outputs figures to visualize the statistics. These figures are saved directly to the `figures/` folder in the data directory in addition to be output such that they can be viewed in a Jupyter notebook. The 3D histogram are generated using the fuction `histogram3d` located in `database_tools/plotting/`.\n\n### Generate Records\nOnce data has been extracted TFRecords can be generated for training a Tensorflow model. The class `RecordsHandler` contains the method `GenerateRecords` which is used to create the TFRecords. This can be done using `scripts/generate_records.py` (paths must be configured). When calling `GenerateRecords` the size of the train, validation, and test splits, as well as the max number of samples per file and a boolean to control whether or not the data is standardized must be specified (using `sklearn.preprocessing.StandardScaler()`.\n\n### Read Records\nThe class `RecordsHandler` also contains the function `ReadRecords` which can be used to read the TFRecords into a Tensorflow `TFRecordsDataset` object. This function can be used to inspect the integrity of the dataset or for loading the dataset for model training. The number of cores and a TensorFlow `AUTOTUNE` object must be provided.\n\n<p align="right">(<a href="#readme-top">back to top</a>)</p>\n\n\n\n<!-- LICENSE -->\n## License\n\nDistributed under the MIT License. See `LICENSE.txt` for more information.\n\n<p align="right">(<a href="#readme-top">back to top</a>)</p>\n',
    'author': 'Cameron Johnson',
    'author_email': 'None',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<3.10',
}


setup(**setup_kwargs)



import itertools
import numpy as np
from neurokit2.ppg.ppg_findpeaks import _ppg_findpeaks_bishop

def detect_flat_lines(x: np.ndarray, n: int) -> bool:
    """
    Flat line detection.

    Args:
        x (np.ndarray): Signal to process.
        n (int): Maximum length of flat line.

    Returns:
        bool: True if n identical values exist consecutively in x.
    """
    return any(sum(1 for _ in g) > (n - 1) for _, g in itertools.groupby(x))

def detect_peaks(sig, show=False, **kwargs):
    """Modified version of neuroki2 ppg_findpeaks method. Returns peaks and troughs
       instead of just peaks. See neurokit2 documentation for original function.
    """
    peaks, troughs = _ppg_findpeaks_bishop(sig, show=show, **kwargs)
    return dict(peaks=peaks[0], troughs=troughs[0])

def detect_notches(sig: np.ndarray, peaks: np.ndarray, troughs: np.ndarray, dx: int = 10, thresh=10) -> list:
    """Detect dichrotic notch by find the maximum velocity
       at least 10 samples after peak and 30 samples before
       the subsequent trough.

    Args:
        sig (np.ndarray): Cardiac signal.
        peaks (list): List of signal peak indices.
        troughs (list): List of signal trough indices.
        dx (int, optional): Spacing between sig values (for np.gradient). Defaults to 10.

    Returns:
        notches (list): List of dichrotic notch indices.
    """
    # always start with first peak
    try:
        if peaks[0] > troughs[0]:
            troughs = troughs[1::]
    except IndexError:
        return []

    notches = []
    for i, j in zip(peaks, troughs):
        try:
            vel = np.gradient(sig[i:j], dx)
            vel_len = vel.shape[0]
            n = np.argmax(vel[int(vel_len / 100 * 25):int(vel_len / 100 * 75)])
            notches.append(n + i)  # add first index of slice to get correct notch index
        except ValueError:  # gradient fails if slice of sig is too small
            continue

    # look for a notch after the last peak if the highest index is a peak.
    try:
        if peaks[-1] > troughs[-1]:
            try:
                vel = np.gradient(sig[peaks[-1]::], dx)
                vel_len = vel.shape[0]
                n = np.argmax(vel[int(vel_len / 100 * 25):int(vel_len / 100 * 75)])
                notches.append(n + peaks[-1])
            except ValueError:
                pass
    except IndexError:
        pass

    # remove notches that are closer than thresh distance in samples
    a, b = np.meshgrid(notches, peaks)
    peak_notch_distances = np.abs(b - a)
    valid_notch_idx = np.array(np.where(peak_notch_distances >= thresh))[1, :]  # get index in notch array
    notches = np.array(notches)[valid_notch_idx]
    return notches
